# command line interface helpers for objects  # noqa: D100
from __future__ import annotations

import logging
import sys

import asyncclick as click

from moat.util import NotGiven

logger = logging.getLogger(__name__)

__all__ = ["std_command"]


class _InvSub:
    """
    This class is a helper for "std_command".
    """

    def __init__(
        self,
        name,
        id_name="name",
        id_typ=str,
        aux=(),
        name_cb=None,
        id_cb=None,
        postproc=None,
        apply=None,
        short_help=None,
        sub_base=None,
        sub_name=None,
        long_name=None,
        prepare=None,
        list_recursive=False,
    ):
        self.name = name
        self.id_name = id_name
        self.id_typ = id_typ
        self.id_cb = id_cb or (lambda _c, _k, x: x)
        self.apply = apply or (lambda _c, _t, _x: None)
        self.name_cb = name_cb or (lambda _c, _k, x: x)
        self.aux = aux
        self.short_help = short_help
        self.postproc = postproc or (lambda _c, _t, _x: None)
        self.long_name = long_name or name
        self.sub_base = sub_base
        self.prepare = prepare
        self.list_recursive = list_recursive
        if sub_name is NotGiven:
            self.sub_name = None
        else:
            self.sub_name = sub_name or name

    def id_arg(self, proc):
        if self.id_name is None:
            return proc
        return click.argument(self.id_name, type=self.id_typ, callback=self.id_cb, nargs=1)(proc)

    def apply_aux(self, proc):
        for t in self.aux:
            proc = t(proc)
        return proc


def std_command(cli, *a, **kw):
    """
    This procedure builds the interface for an inventory-ish thing.

    Usage::

        @click.group(short_help="Manage computer inventory.")
        @click.pass_obj
        async def cli(obj):
            "Inventorize your computers, networks, and their connections."
            obj.inv = await InventoryRoot.as_handler(obj.client)

        std_command(
            cli,  # from above
            "vlan",  # subcommand name
            "id",  # item identifier
            int,  # item identifier type
            aux=(  # additional attributes
                click.option("-d", "--desc", type=str, default=None, help="Description"),
                click.option("-w", "--wlan", type=str, default=None, help="WLAN SSID"),
            ),
            short_help="Manage VLANs",
        )
    """
    tinv = _InvSub(*a, **kw)
    tname = tinv.name
    tnname = "n_" + tname

    def this(obj):
        # Delayed resolving of the actual thing subhierarchy
        if tinv.sub_base:
            data = getattr(obj, tinv.sub_base)
        else:
            data = obj.data
        if tinv.sub_name:
            return getattr(data, tinv.sub_name)
        return data

    @cli.group(
        name=tname,
        invoke_without_command=True,
        short_help=tinv.short_help,
        help=f"""\
            Manager for {tinv.long_name}s.

            \b
            Use '… {tname} -' to list all entries.
            Use '… {tname} NAME' to show details of a single entry.
            """,
    )
    @click.argument("name", type=str, nargs=1)
    @click.pass_context
    async def typ(ctx, name):
        obj = ctx.obj
        if tinv.prepare is not None:
            await tinv.prepare(obj)

        if name == "-":
            if ctx.invoked_subcommand is not None:
                raise click.BadParameter("The name '-' triggers a list and precludes subcommands.")
            cnt = 0
            for n in this(obj).all_children if tinv.list_recursive else this(obj):
                cnt += 1
                print(n, file=obj.stdout)
            if not cnt and ctx.obj.debug:
                print("no entries", file=sys.stderr)
        elif ctx.invoked_subcommand is None:
            # Show data from a single entry
            n = this(obj).by_name(name)
            if n is None:
                raise KeyError(name)
            cnt = 0
            for k in n.ATTRS + getattr(n, "AUX_ATTRS", ()):
                v = getattr(n, k, None)
                if v is not None:
                    cnt += 1
                    if isinstance(v, dict):
                        v = v.items()
                    if isinstance(v, type({}.items())):  # pylint: disable=W1116
                        for kk, vv in sorted(v):
                            if isinstance(vv, (tuple, list)):
                                if vv:
                                    vv = " ".join(str(x) for x in vv)  # noqa:PLW2901
                                else:
                                    vv = "-"  # noqa:PLW2901
                            elif isinstance(vv, dict):
                                vv = " ".join(f"{x}={y}" for x, y in sorted(vv.items()))  # noqa:PLW2901
                            print(f"{k} {kk} {vv}", file=obj.stdout)
                    else:
                        print(f"{k} {v}", file=obj.stdout)
            if not cnt and ctx.obj.debug:
                print("exists, no data", file=sys.stderr)
        else:
            obj[tnname] = name
            try:
                obj[tname] = this(obj).by_name(name)
            except KeyError:
                obj[tname] = None

    def alloc(obj, name):
        # Allocate a new thing
        if isinstance(name, (tuple, list)):
            n = this(obj).follow(name, create=True)
        else:
            n = this(obj).allocate(name)
        return n

    # This is a hack for groups that take arguments. Doesn't work for
    # multiple arguments, but you shouldn't do that anyway.
    @typ.command("--help", hidden=True)
    @click.pass_context
    def help_(ctx):  # pylint:disable=unused-variable  # oh boy
        print(typ.get_help(ctx))

    @typ.command(short_help="Add a " + tinv.long_name)
    @tinv.id_arg
    @tinv.apply_aux
    @click.pass_obj
    async def add(obj, **kw):
        name = obj[tnname]
        if obj[tname] is not None:
            raise RuntimeError(f"{name} already exists")
        if tinv.id_name:
            kw["name"] = name
            n = alloc(obj, kw.pop(tinv.id_name))
        else:
            n = alloc(obj, name)

        await _v_mod(obj, n, **kw)

    add.__doc__ = f"""
        Add a {tinv.long_name}
        """

    @typ.command("set", short_help="Modify a " + tinv.long_name)
    @tinv.apply_aux
    @click.pass_obj
    async def set_(obj, **kw):
        n = obj[tname]
        if n is None:
            raise KeyError(tname)

        await _v_mod(obj, n, **kw)

    set_.__doc__ = f"""
        Modify a {tinv.long_name}
        """

    @typ.command(short_help="Delete a " + tinv.long_name)
    @click.pass_obj
    async def delete(obj, **kw):  # noqa: ARG001
        name = obj[tnname]
        n = this(obj).by_name(name)
        if n is not None:
            await n.delete()

    delete.__doc__ = """
        Delete a {tinv.long_name}
        """

    async def _v_mod(obj, thing, **kw):
        tinv.apply(obj, thing, kw)
        for k, v in kw.items():
            if v:
                if v == "-":
                    v = None  # noqa:PLW2901
                try:
                    setattr(thing, k, v)
                except AttributeError:
                    if k != "name":
                        raise AttributeError(k, v) from None
        tinv.postproc(obj, thing, kw)
        await thing.save()

    # Finally, return the CLI so the user can attach more stuff
    return typ



"""
This module locally stores code that's been uploaded to MoaT-KV
so that it can be called easily.

"Code" consists of either Python modules or single procedures.

"""

from __future__ import annotations

import anyio
import contextlib
import logging
import sys
from functools import partial

from moat.util import NotGiven, P, make_module, make_proc

from .obj import ClientEntry, ClientRoot

logger = logging.getLogger(__name__)


class EmptyCode(Exception):
    """
    There is no code here.
    """

    pass


class ModuleRoot(ClientRoot):
    """
    Modules are stored at ``(*PREFIX,*name)``.
    The prefix defaults to ``("code","module")``.

    The module code is stored textually. Content that is not UTF-8 is TODO.
    """

    CFG = "modules"

    err: ErrorRoot = None  # noqa: F821

    @classmethod
    def child_type(cls, name):  # noqa:D102,ARG003
        return ModuleEntry

    async def run_starting(self):  # noqa:D102
        from .errors import ErrorRoot  # noqa: PLC0415

        self.err = await ErrorRoot.as_handler(self.client)
        await super().run_starting()

    async def add(self, path, code=None):
        """
        Add or replace this code at this location.
        """
        # test-compile the code for validity
        if code is None:
            return await self.remove(path)

        make_module(code, path)

        r = await self.client.set(self._path + path, value=dict(code=code), nchain=2)
        await self.wait_chain(r.chain)

    async def remove(self, path):
        """
        Remove code at this location.
        """
        entry = self.follow(self._path + path, create=False)
        return await entry.delete()


class ModuleEntry(ClientEntry):  # noqa:D101
    _module = None

    @property
    def name(self):  # noqa:D102
        return ".".join(self.subpath)

    async def set_value(self, value):  # noqa:D102
        await super().set_value(value)
        if value is NotGiven:
            self._module = None
            with contextlib.suppress(KeyError):
                del sys.modules[self.name]
            return

        try:
            c = self.value.code
            if not isinstance(c, str):
                raise TypeError("Not a string, cannot compile")
            m = make_module(c, self.subpath)
        except Exception as exc:
            self._module = None
            logger.warning("Could not compile @%r", self.subpath)
            await self.root.err.record_error(
                "compile",
                self.subpath,
                exc=exc,
                message="compiler error",
            )
        else:
            await self.root.err.record_working("compile", self.subpath)
            self._module = m


class CodeRoot(ClientRoot):
    """
    This class represents the root of a code storage hierarchy. Ideally
    there should only be one, but you can configure more.

    You typically don't create this class directly; instead, call
    :meth:`CodeRoot.as_handler`::

        code = await CodeRoot.as_handler(client, your_config.get("code-storage",{})

    The prefix defaults to ``("code","proc")``.

    Configuration:

    Arguments:
      prefix (list): Where to store the code in MoaT-KV.
        The default is ``('.moat','kv','code','proc')``.

    The code is stored as a dict.

    Arguments:
      code: the actual code to run. It is parsed as a function body; be
        aware that multi-line strings will be indented more than you'd like.
      async: flag whether the code should run asynchronously.
        True: yes, None (default): no, False: run in a separate thread.
      vars: array of names to be used as named arguments.

    All arguments are mandatory and should be named.
    Extra arguments will be available in the "kw" dict.

    """

    CFG = "codes"

    err = None

    @classmethod
    def child_type(cls, name):  # noqa:D102,ARG003
        return CodeEntry

    async def run_starting(self):  # noqa:D102
        from .errors import ErrorRoot  # noqa: PLC0415

        self.err = await ErrorRoot.as_handler(self.client)
        await super().run_starting()

    async def add(self, path, code=None, *, is_async=None, variables=()):
        """
        Add or replace this code at this location.
        """
        if code is NotGiven:
            return await self.remove(path)

        # test-compile the code for validity
        make_proc(code, variables, path, use_async=is_async)

        r = await self.client.set(
            self._path + path,
            value=dict(code=code, is_async=is_async, vars=variables),
            nchain=2,
        )
        await self.wait_chain(r.chain)

    async def remove(self, path):
        """Drop this code"""
        r = await self.client.set(*self._path, *path, value=None, nchain=2)
        await self.wait_chain(r.chain)

    def __call__(self, name, *a, **kw):
        """Call a (named) procedure below this.

        Return: whatever it returns.
        """
        if isinstance(name, str):
            name = name.split(".")
        c = self
        for k in name:
            c = c[k]
        if c is self:
            raise RuntimeError("Empty code names don't make sense")
        return c(*a, **kw)


class CodeEntry(ClientEntry):  # noqa:D101
    _code = None
    is_async = None

    def __init__(self, *a, **kv):
        self.reload_event = anyio.Event()
        super().__init__(*a, *kv)

    @property
    def name(self):  # noqa:D102
        return P(self.subpath)

    async def set_value(self, value):  # noqa:D102
        await super().set_value(value)
        if value is NotGiven:
            self._code = None
            return

        try:
            v = self.value
            c = v["code"]
            a = v.get("is_async", None)
            p = make_proc(c, v.get("vars", ()), self.subpath, use_async=a)
        except Exception as exc:
            logger.warning("Could not compile @%s", self.subpath)
            await self.root.err.record_error(
                "compile",
                self.subpath,
                exc=exc,
                message="compiler error",
            )
            self._code = None
        else:
            await self.root.err.record_working("compile", self.subpath)
            self._code = p
            self.is_async = a
            self.reload_event.set()
            self.reload_event = anyio.Event()

    def __call__(self, *a, **kw):
        """Call this code block.

        Return: whatever it returns.
        """
        if self._code is None:
            raise EmptyCode(self._path)
        if self.is_async is False:
            proc = self._code
            if kw:
                proc = partial(proc, **kw)
            return anyio.to_thread.run_sync(proc, *a)

        return self._code(*a, **kw)



#
"""
Password-based auth method.

Does not limit anything, allows everything.
"""

from __future__ import annotations

import contextlib

import nacl.secret

from moat.kv.client import Client, NoData
from moat.kv.exceptions import AuthFailedError

from . import (
    BaseClientAuth,
    BaseClientAuthMaker,
    BaseServerAuthMaker,
    RootServerUser,
    null_client_login,
    null_server_login,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.kv.model import Entry
    from moat.kv.server import StreamCommand


def load(typ: str, *, make: bool = False, server: bool):  # noqa:D103
    if typ == "client":
        if server:
            return null_server_login
        else:
            return null_client_login
    if typ != "user":
        raise NotImplementedError("This module only handles users")
    if server:
        if make:
            return ServerUserMaker
        else:
            return ServerUser
    else:
        if make:
            return ClientUserMaker
        else:
            return ClientUser


async def pack_pwd(client, password, length):
    """Client side: encrypt password"""
    secret = await client.dh_secret(length=length)
    from hashlib import sha256  # noqa: PLC0415

    pwd = sha256(password).digest()
    box = nacl.secret.SecretBox(secret)
    pwd = box.encrypt(pwd)
    return pwd


async def unpack_pwd(client, password):
    """Server side: extract password"""
    box = nacl.secret.SecretBox(client.dh_key)
    if isinstance(password, memoryview):
        password = bytes(password)
    pwd = box.decrypt(password)
    return pwd
    # TODO check with Argon2


class ServerUserMaker(BaseServerAuthMaker):  # noqa:D101
    _name = None
    _aux = None
    password: str = None

    @property
    def ident(self):  # noqa:D102
        return self._name

    @classmethod
    async def recv(cls, cmd, data):  # noqa:D102
        self = cls()
        self._name = data["ident"]
        self._aux = data.get("aux")
        pwd = data.get("password")
        pwd = await unpack_pwd(cmd.client, pwd)

        # TODO use Argon2 to re-hash this
        self.password = pwd
        return self

    async def send(self, cmd):  # noqa:D102,ARG002
        return  # nothing to do, we don't share the hash

    @classmethod
    def load(cls, data):  # noqa:D102
        self = super().load(data)
        self._name = data.path[-1]
        return self

    def save(self):  # noqa:D102
        res = super().save()
        res["password"] = self.password
        return res


class ServerUser(RootServerUser):  # noqa:D101
    @classmethod
    def load(cls, data: Entry):
        """Create a ServerUser object from existing stored data"""
        self = super().load(data)
        self._name = data.name
        return self

    async def auth(self, cmd: StreamCommand, data):
        """Verify that @data authenticates this user."""
        await super().auth(cmd, data)

        pwd = await unpack_pwd(cmd.client, data.password)
        if pwd != self.password:  # pylint: disable=no-member
            # pylint: disable=no-member
            raise AuthFailedError("Password hashes do not match", self._name)


class ClientUserMaker(BaseClientAuthMaker):  # noqa:D101
    gen_schema = dict(
        type="object",
        additionalProperties=True,
        properties=dict(
            name=dict(type="string", minLength=1, pattern="^[a-zA-Z][a-zA-Z0-9_]*$"),
            password=dict(type="string", minLength=5),
        ),
        required=["name", "password"],
    )
    mod_schema = dict(
        type="object",
        additionalProperties=True,
        properties=dict(password=dict(type="string", minLength=5)),
        # required=[],
    )
    _name = None
    _pass = None
    _length = 1024

    @property
    def ident(self):  # noqa:D102
        return self._name

    # Overly-complicated methods of exchanging the user name

    @classmethod
    def build(cls, user, _initial=True):  # noqa:D102
        self = super().build(user, _initial=_initial)
        self._name = user["name"]
        if "password" in user:
            self._pass = user["password"].encode("utf-8")
        return self

    @classmethod
    async def recv(cls, client: Client, ident: str, _kind: str = "user", _initial=True):
        """Read a record representing a user from the server."""
        m = await client._request(  # noqa:SLF001
            action="auth_get",
            typ=cls._auth_method,
            kind=_kind,
            ident=ident,
            nchain=0 if _initial else 2,
        )
        # just to verify that the user exists
        # There's no reason to send the password hash back
        self = cls(_initial=_initial)
        self._name = m.name
        with contextlib.suppress(AttributeError):
            self._chain = m.chain
        return self

    async def send(self, client: Client, _kind="user", **msg):  # pylint: disable=unused-argument,arguments-differ
        """Send a record representing this user to the server."""
        if self._pass is not None:
            msg["password"] = await pack_pwd(client, self._pass, self._length)

        await client._request(  # noqa:SLF001
            action="auth_set",
            ident=self._name,
            typ=type(self)._auth_method,  # noqa:SLF001
            kind=_kind,
            chain=self._chain,
            **msg,
        )

    def export(self):
        """Return the data required to re-create the user via :meth:`build`."""
        res = super().export()
        res["name"] = self._name
        return res


class ClientUser(BaseClientAuth):  # noqa:D101
    schema = dict(
        type="object",
        additionalProperties=True,
        properties=dict(
            name=dict(type="string", minLength=1, pattern="^[a-zA-Z][a-zA-Z0-9_]*$"),
            password=dict(type="string", minLength=5),
        ),
        required=["name", "password"],
    )
    _name = None
    _pass = None
    _length = 1024

    @property
    def ident(self):  # noqa:D102
        return self._name

    @classmethod
    def build(cls, user):  # noqa:D102
        self = super().build(user)
        self._name = user["name"]
        self._pass = user["password"].encode("utf-8")
        return self

    async def auth(self, client: Client, chroot=()):
        """
        Authorizes this user with the server.
        """
        chroot  # noqa:B018
        try:
            pw = await pack_pwd(client, self._pass, self._length)
            await client._request(  # noqa:SLF001
                action="auth",
                typ=self._auth_method,
                iter=False,
                ident=self.ident,
                password=pw,
                **self.auth_data(),
            )
        except NoData:
            pass



"""
This module implements a :class:`asyncactor.Actor` which works on top of
a MoaT-KV client.
"""

from __future__ import annotations

from asyncactor import Actor
from asyncactor.abc import MonitorStream, Transport

__all__ = [
    "ActorState",
    "BrokenState",
    "ClientActor",
    "CompleteState",
    "DetachedState",
    "PartialState",
]


class ClientActor(Actor):
    "An actor that runs over a `ClientTranspport`"

    def __init__(self, client, *a, topic, **kw):
        super().__init__(ClientTransport(client, topic), *a, **kw)


class ClientTransport(Transport):
    """
    This class exports the client's direct messaging interface to the
    actor.
    """

    def __init__(self, client, topic):
        self.client = client
        self.topic = topic

    def monitor(self):
        return ClientMonitor(self)

    async def send(self, payload):
        await self.client.msg_send(self.topic, payload)


class ClientMonitor(MonitorStream):
    _mon1 = None
    _mon2 = None
    _it = None

    async def __aenter__(self):
        self._mon1 = self.transport.client.msg_monitor(self.transport.topic)
        self._mon2 = await self._mon1.__aenter__()
        return self

    async def __aexit__(self, *tb):
        return await self._mon1.__aexit__(*tb)

    def __aiter__(self):
        self._it = self._mon2.__aiter__()
        return self

    async def __anext__(self):
        msg = await self._it.__anext__()
        return msg.data


# The following events are used by Runner etc. to notify running jobs
# about the current connectivity state.
#
class ActorState:
    """base class for states"""

    def __init__(self, msg=None):
        self.msg = msg

    def __repr__(self):
        return f"<{self.__class__.__name__}:{self.msg!r}>"


class BrokenState(ActorState):
    """I have no idea what's happening, probably nothing good"""

    pass


class DetachedState(ActorState):
    """I am detached, my actor group is not visible"""

    pass


class PartialState(ActorState):
    """Some but not all members of my actor group are visible"""

    pass


class CompleteState(ActorState):
    """All members of my actor group are visible"""

    pass



"""
Data access
"""

from __future__ import annotations

import contextlib
import os
import sys
import time

from moat.util import NotGiven, Path, attrdict, process_args, yprint
from moat.util.times import ts2iso

from collections.abc import Mapping


def add_dates(d):
    """
    Given a dict with int/float entries that might conceivably be dates,
    add ``_*`` with a textual representation.
    """

    t = time.time()
    start = t - 366 * 24 * 3600
    stop = t + 366 * 24 * 3600

    def _add(d):
        if isinstance(d, (list, tuple)):
            for dd in d:
                _add(dd)
            return
        if not isinstance(d, Mapping):
            return
        for k, v in list(d.items()):
            if isinstance(k, str) and k.startswith("_"):
                continue
            if not isinstance(v, (int, float)):
                _add(v)
                continue
            if start <= v <= stop:
                d[f"_{k}"] = ts2iso(v)

    _add(d)


async def data_get(
    conn,
    path,
    *,
    meta=False,
    recursive=True,
    as_dict="_",
    maxdepth=-1,
    mindepth=0,
    empty=False,
    raw=False,
    internal=False,
    path_mangle=None,
    item_mangle=None,
    add_date=False,
    out=None,
):
    """Generic code to dump a subtree.

    `path_mangle` accepts a path and the as_dict parameter. It should
    return the new path. This is used for e.g. prefixing the path with a
    device name. Returning ``None`` causes the entry to be skipped.
    """
    if path_mangle is None:

        def path_mangle(x):
            return x

    if item_mangle is None:

        async def item_mangle(x):  # pylint: disable=function-redefined
            return x

    if out is None:
        out = sys.stdout
    elif out is False:
        out = []

    if recursive:
        kw = {}
        if maxdepth is not None and maxdepth >= 0:
            kw["max_depth"] = maxdepth
        if mindepth:
            kw["min_depth"] = mindepth
        if empty:
            kw["empty"] = True
        if meta:
            kw.setdefault("nchain", meta)
        y = {}
        if internal:
            res = await conn._request(action="get_tree_internal", path=path, iter=True, **kw)  # noqa:SLF001
        else:
            res = conn.get_tree(path, **kw)
        async for r in res:
            r = await item_mangle(r)  # noqa:PLW2901
            if r is None:
                continue
            r.pop("seq", None)
            path = r.pop("path")
            path = path_mangle(path)
            if path is None:
                continue
            if add_date and "value" in r:
                add_dates(r.value)

            if as_dict is not None:
                yy = y
                for p in path:
                    yy = yy.setdefault(p, {})
                try:
                    yy[as_dict] = r if meta else r.value
                except AttributeError:
                    if empty:
                        yy[as_dict] = None
            else:
                if raw:
                    y = path
                else:
                    y = {}
                    try:
                        y[path] = r if meta else r.value
                    except AttributeError:
                        if empty:
                            y[path] = None
                        else:
                            continue

                if isinstance(out, list):
                    out.append(y)
                else:
                    yprint([y], stream=out)

        if as_dict is not None:
            if maxdepth:

                def simplex(d):
                    for k, v in d.items():
                        if isinstance(v, dict):
                            d[k] = simplex(d[k])
                    if as_dict in d and d[as_dict] is None:
                        if len(d) == 1:
                            return None
                        else:
                            del d[as_dict]
                    return d

                y = simplex(y)
            if isinstance(out, list):
                return y
            yprint(y, stream=out)
        return  # end "if recursive"

    res = await conn.get(path, nchain=meta)
    if add_date and "value" in res:
        add_dates(res.value)
    if not meta:
        try:
            res = res.value
        except AttributeError:
            #           if obj.debug:
            #               print("No data at", path, file=sys.stderr)
            return

    if out is False:
        return res
    if not raw:
        yprint(res, stream=out)
    elif isinstance(res, bytes):
        os.write(out.fileno(), res)
    else:
        out.write(str(res))
    pass  # end get


def res_get(res, attr: Path, **kw):  # pylint: disable=redefined-outer-name
    """
    Get a node's value and access the dict items beneath it.

    The node value must be an attrdict.
    """
    val = res.get("value", None)
    if val is None:
        return None
    return val._get(attr, **kw)  # noqa:SLF001


def res_update(res, attr: Path, value=None, **kw):  # pylint: disable=redefined-outer-name
    """
    Set a node's sub-item's value, possibly merging dicts.
    Entries set to 'NotGiven' are deleted.

    The node value must be an attrdict.

    Returns the new value.
    """
    val = res.get("value", attrdict())
    return val._update(attr, value=value, **kw)  # noqa:SLF001


async def node_attr(obj, path, res=None, chain=None, **kw):
    """
    Sub-attr setter.

    Args:
        obj: command object
        path: address of the node to change
        res: old node, if it has been read already
        chain: change chain of node, copied from res if clear
        **kw: the results of `attr_args`

    Returns the result of setting the attribute.
    """
    if res is None:
        res = await obj.client.get(path, nchain=obj.meta or 2)
    if chain is None:
        with contextlib.suppress(AttributeError):
            chain = res.chain
    try:
        val = res.value
    except AttributeError:
        chain = None
        val = NotGiven
    val = process_args(val, **kw)
    if val is NotGiven:
        res = await obj.client.delete(path, nchain=obj.meta, chain=chain)
    else:
        res = await obj.client.set(path, value=val, nchain=obj.meta, chain=chain)
    return res



# command line interface  # noqa:D100
from __future__ import annotations

import contextlib

import asyncclick as click
from range_set import RangeSet

from moat.util import P, PathLongener, yprint

from collections.abc import Mapping


@click.group(short_help="Control internal state.")  # pylint: disable=undefined-variable
async def cli():
    """
    This subcommand queries and controls the server's internal state.
    """
    pass


@cli.command()
@click.option("-n", "--nodes", is_flag=True, help="Get node status.")
@click.option("-d", "--deleted", is_flag=True, help="Get deletion status.")
@click.option("-m", "--missing", is_flag=True, help="Get missing-node status.")
@click.option(
    "-r",
    "--remote-missing",
    "remote_missing",
    is_flag=True,
    help="Get remote-missing-node status.",
)
@click.option("-p", "--present", is_flag=True, help="Get known-data status.")
@click.option("-s", "--superseded", is_flag=True, help="Get superseded-data status.")
@click.option("-D", "--debug", is_flag=True, help="Get internal verbosity.")
@click.option("--debugger", is_flag=True, help="Start a remote debugger. DO NOT USE.")
@click.option("-k", "--known", hidden=True, is_flag=True, help="Get superseded-data status.")
@click.option("-a", "--all", is_flag=True, help="All available data.")
@click.pass_obj
async def state(obj, **flags):
    """
    Dump the server's state.
    """
    if flags.pop("known", None):
        flags["superseded"] = True
    if flags.pop("all", None):
        flags["superseded"] = True
        flags["present"] = True
        flags["nodes"] = True
        flags["deleted"] = True
        flags["missing"] = True
        flags["remote_missing"] = True
    res = await obj.client._request("get_state", iter=False, **flags)  # noqa:SLF001
    k = res.pop("known", None)
    if k is not None:
        res["superseded"] = k
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-d", "--deleted", is_flag=True, help="Mark as deleted. Default: superseded")
@click.option(
    "-n",
    "--node",
    "source",
    default="?",
    help="The node this message is faked as being from.",
)
@click.option("-b", "--broadcast", is_flag=True, help="Send to all servers")
@click.argument("node", nargs=1)
@click.argument("items", type=int, nargs=-1)
@click.pass_obj
async def mark(obj, deleted, source, node, items, broadcast):
    """
    Fix internal state. Use no items to fetch the current list from the
    server's ``missing`` state. Use an empty node name to add the whole
    list, not just a single node's.

    This is a dangerous command.
    """

    k = "deleted" if deleted else "superseded"
    if not items:
        r = await obj.client._request("get_state", iter=False, missing=True)  # noqa:SLF001
        r = r["missing"]
        if node != "":
            r = {node: r[node]}
    elif node == "":
        raise click.UsageError("You can't do that with an empty node")
    else:
        r = RangeSet()
        for i in items:
            r.add(i)
        r = {node: r.__getstate__()}

    msg = {k: r, "node": source}

    await obj.client._request("fake_info", iter=False, **msg)  # noqa:SLF001
    if broadcast:
        await obj.client._request("fake_info_send", iter=False, **msg)  # noqa:SLF001

    res = await obj.client._request("get_state", iter=False, **{k: True})  # noqa:SLF001
    yprint(res, stream=obj.stdout)


@cli.command(short_help="Manage the Deleter list")
@click.option("-d", "--delete", is_flag=True, help="Remove these nodes")
@click.argument("nodes", nargs=-1)
@click.pass_obj
async def deleter(obj, delete, nodes):
    """
    Manage the Deleter list

    This is the set of nodes that must be online for removal of deleted
    entries from MoaT-KV's data.

    There should be one such node in every possible network partition.
    Also, all nodes with permanent storage should be on the list.

    Usage:
    - … deleter          -- list state
    - … deleter NODE…    -- add this node
    - … deleter -d NODE… -- remove this node
    """

    res = await obj.client._request(  # noqa:SLF001
        action="get_internal",
        path=("actor", "del"),
        iter=False,
        nchain=3 if delete or nodes else 2,
    )
    val = res.get("value", {})
    if isinstance(val, Mapping):
        val = val.get("nodes", [])
    # else: compatibility, TODO remove
    val = set(val)
    if delete:
        if nodes:
            val -= set(nodes)
        else:
            val = set()
    elif nodes:
        val |= set(nodes)
    else:
        yprint(res, stream=obj.stdout)
        return

    val = {"nodes": list(val)}
    res = await obj.client._request(  # noqa:SLF001
        action="set_internal",
        path=("actor", "del"),
        iter=False,
        chain=res.chain,
        value=val,
    )
    res.value = val
    yprint(res, stream=obj.stdout)


@cli.command()
@click.argument("path", nargs=1)
@click.pass_obj
async def dump(obj, path):
    """
    Dump internal state.

    This displays MoaT-KV's internal state.
    """

    path = P(path)
    y = {}
    pl = PathLongener()
    async for r in await obj.client._request("get_tree_internal", path=path, iter=True, nchain=0):  # noqa:SLF001
        pl(r)
        path = r["path"]
        yy = y
        for p in path:
            yy = yy.setdefault(p, {})
        with contextlib.suppress(KeyError):
            yy["_"] = r["value"]
    yprint(y, stream=obj.stdout)


@cli.command()
@click.argument("node", nargs=1)
@click.argument("tick", type=int, nargs=1)
@click.pass_obj
async def get(obj, node, tick):
    """
    Fetch data by node+tick.

    This looks up internal data.
    """

    res = await obj.client._request("get_value", node=node, tick=tick, nchain=99)  # noqa:SLF001
    if not obj.meta:
        res = res.value
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-n", "--num", type=int, help="Return at most this many IDs")
@click.option("-c", "--current", is_flag=True, help="Return only IDs with current data")
@click.option("-C", "--copy", is_flag=True, help="Create an no-op change")
@click.argument("node", nargs=1)
@click.pass_obj
async def enum(obj, node, num, current, copy):
    """
    List IDs of live data by a specific node.

    Can be used to determine whether a node still has live data,
    otherwise it can be deleted.

    If '--current' is set, only the IDs of the entries that have last been
    updated by that node are shown. '--copy' rewrites these entries.

    Increase verbosity to also show the oject paths.
    """

    res = await obj.client._request("enum_node", node=node, max=num, current=current)  # noqa:SLF001
    if obj.meta and not copy and obj.debug <= 1:
        yprint(res, stream=obj.stdout)
    else:
        for k in res.result:
            if copy or obj.debug > 1:
                res = await obj.client._request("get_value", node=node, tick=k, nchain=3)  # noqa:SLF001
                if obj.debug > 1:
                    print(k, res.path)
                else:
                    print(k)
                if copy and res.chain.node == node:
                    res = await obj.client.set(res.path, value=res.value, chain=res.chain)
            else:
                print(k)


@cli.command()
@click.argument("node", nargs=1)
@click.pass_obj
async def kill(obj, node):
    """
    Remove a node from the node list.

    This command only works if this node does not have any current data in
    the system.
    """
    res = await obj.client._request("kill_node", node=node)  # noqa:SLF001
    if obj.meta:
        yprint(res, stream=obj.stdout)



# command line interface  # noqa:D100
from __future__ import annotations

import asyncclick as click

from moat.util import NotGiven, P, Path, PathLongener, yload, yprint


@click.group()  # pylint: disable=undefined-variable
async def cli():
    """Manage codecs and converters. Usage: … codec …"""
    pass


@cli.command()
@click.option("-e", "--encode", type=click.File(mode="w", lazy=True), help="Save the encoder here")
@click.option("-d", "--decode", type=click.File(mode="w", lazy=True), help="Save the decoder here")
@click.option("-s", "--script", type=click.File(mode="w", lazy=True), help="Save the data here")
@click.argument("path", nargs=1)
@click.pass_obj
async def get(obj, path, script, encode, decode):
    """Read type information"""
    path = P(path)
    if not len(path):
        raise click.UsageError("You need a non-empty path.")
    res = await obj.client._request(  # noqa:SLF001
        action="get_internal",
        path=Path("codec") + path,
        iter=False,
        nchain=obj.meta,
    )
    if encode and res.get("encode", None) is not None:
        encode.write(res.pop("encode"))
    if decode and res.get("decode", None) is not None:
        decode.write(res.pop("decode"))

    if not obj.meta:
        res = res.value
    yprint(res, stream=script or obj.stdout)


@cli.command(name="list")
@click.pass_obj
@click.argument("path", nargs=1)
async def list_(obj, path):
    """List type information entries"""
    res = await obj.client._request(  # noqa:SLF001
        action="get_tree_internal",
        path=Path("codec") + P(path),
        iter=True,
        nchain=obj.meta,
    )
    pl = PathLongener(())
    async for r in res:
        pl(r)
        print(" ".join(str(x) for x in r.path), file=obj.stdout)


@cli.command("set")
@click.option("-e", "--encode", type=click.File(mode="r"), help="File with the encoder")
@click.option("-d", "--decode", type=click.File(mode="r"), help="File with the decoder")
@click.option("-D", "--data", type=click.File(mode="r"), help="File with the rest")
@click.option("-i", "--in", "in_", nargs=2, multiple=True, help="Decoding sample")
@click.option("-o", "--out", nargs=2, multiple=True, help="Encoding sample")
@click.argument("path", nargs=1)
@click.pass_obj
async def set_(obj, path, encode, decode, data, in_, out):
    """Save codec information"""
    path = P(path)
    if not len(path):
        raise click.UsageError("You need a non-empty path.")

    if data:
        msg = yload(data)
    else:
        msg = {}
    chain = NotGiven
    if "value" in msg:
        chain = msg.get("chain", NotGiven)
        msg = msg["value"]

    if "encode" in msg:
        if encode:
            raise click.UsageError("Duplicate encode script")
    else:
        if not encode:
            raise click.UsageError("Missing encode script")
        msg["encode"] = encode.read()
    if "decode" in msg:
        if decode:
            raise click.UsageError("Duplicate decode script")
    else:
        if not decode:
            raise click.UsageError("Missing decode script")
        msg["decode"] = decode.read()
    if in_:
        msg["in"] = [(eval(a), eval(b)) for a, b in in_]  # pylint: disable=eval-used
    if out:
        msg["out"] = [(eval(a), eval(b)) for a, b in out]  # pylint: disable=eval-used

    if not msg["in"]:
        raise click.UsageError("Missing decode tests")
    if not msg["out"]:
        raise click.UsageError("Missing encode tests")

    res = await obj.client._request(  # noqa:SLF001
        action="set_internal",
        value=msg,
        path=Path("codec") + path,
        iter=False,
        nchain=obj.meta,
        **({} if chain is NotGiven else {"chain": chain}),
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-c", "--codec", type=P, help="Codec to link to.")
@click.option("-d", "--delete", is_flag=True, help="Use to delete this converter.")
@click.option(
    "-l",
    "--list",
    "list_",
    is_flag=True,
    help="Use to list this converter; '-' to list all.",
)
@click.argument("name", nargs=1)
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def convert(obj, path, codec, name, delete, list_):
    """Match a codec to a path (read, if no codec given)"""
    path = P(path)
    if delete and list_:
        raise click.UsageError("You can't both list and delete a path.")
    if not len(path) and not list_:
        raise click.UsageError("You need a non-empty path.")
    if codec and delete:
        raise click.UsageError("You can't both set and delete a path.")

    if list_:
        if name == "-":
            if len(path):
                raise click.UsageError("You can't use a path here.")
            res = await obj.client._request(  # noqa:SLF001
                action="enum_internal",
                path=Path("conv"),
                iter=False,
                nchain=0,
                empty=True,
            )
            for r in res.result:
                print(r, file=obj.stdout)

        else:
            res = await obj.client._request(  # noqa:SLF001
                action="get_tree_internal",
                path=Path("conv", name) + path,
                iter=True,
                nchain=obj.meta,
            )
            pl = PathLongener(())
            async for r in res:
                pl(r)
                try:
                    print(f"{r.path} : {Path.build(r.value['codec'])}", file=obj.stdout)
                except Exception as e:
                    print(f"{Path(r.path)} {e!r}", file=obj.stdout)

        return
    if delete:
        res = await obj.client._request(action="delete_internal", path=Path("conv", name) + path)  # noqa:SLF001
    else:
        msg = {"codec": codec}
        res = await obj.client._request(  # noqa:SLF001
            action="set_internal",
            value=msg,
            path=Path("conv", name) + path,
            iter=False,
            nchain=obj.meta,
        )
    if obj.meta:
        yprint(res, stream=obj.stdout)
    elif type or delete:
        print(res.tock, file=obj.stdout)
    else:
        print(" ".join(res.type), file=obj.stdout)



# noqa:D100
from __future__ import annotations

import anyio
import copy
import logging
import time
from contextlib import AsyncExitStack, asynccontextmanager, suppress
from functools import partial
from unittest import mock

import trio
from asyncscope import main_scope, scope

from moat.util import NotGiven, attrdict, combine_dict
from moat.kv.mock import S
from moat.kv.server import Server
from moat.mqtt.broker import create_broker

from . import CFG

logger = logging.getLogger(__name__)

otm = time.time


@asynccontextmanager
async def stdtest(n=1, run=True, ssl=False, tocks=20, **kw):  # noqa:D103
    C_OUT = CFG.get("_stdout", NotGiven)
    if C_OUT is not NotGiven:
        del CFG["_stdout"]
    TESTCFG = copy.deepcopy(CFG["kv"])
    TESTCFG.server.port = None
    TESTCFG.root = "test"
    if C_OUT is not NotGiven:
        TESTCFG["_stdout"] = C_OUT

    from anyio.pytest_plugin import FreePortFactory  # noqa: PLC0415
    from socket import SOCK_STREAM  # noqa: PLC0415

    PORT = FreePortFactory(SOCK_STREAM)()
    broker_cfg = {
        "listeners": {"default": {"type": "tcp", "bind": f"127.0.0.1:{PORT}"}},
        "timeout-disconnect-delay": 2,
        "auth": {"allow-anonymous": True, "password-file": None},
    }
    URI = f"mqtt://127.0.0.1:{PORT}/"

    if ssl:
        import ssl  # noqa:PLC0415,I001
        import trustme  # noqa: PLC0415

        ca = trustme.CA()
        cert = ca.issue_server_cert("127.0.0.1")
        server_ctx = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        client_ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        ca.configure_trust(client_ctx)
        cert.configure_cert(server_ctx)
    else:
        server_ctx = client_ctx = False

    clock = trio.lowlevel.current_clock()
    with suppress(Exception):
        clock.autojump_threshold = 0.02  # networking

    async def mock_get_host_port(st, host):
        i = int(host[host.rindex("_") + 1 :])
        s = st.s[i]
        await s.is_serving
        for host, port, *_ in s.ports:
            if host == "::" or host[0] != ":":
                return host, port

    def tm():
        try:
            return trio.current_time()
        except RuntimeError:
            return otm()

    async def mock_set_tock(self, old):
        assert self._tock < tocks, "Test didn't terminate. Limit:" + str(tocks)
        await old()

    done = False
    async with main_scope("moat.kv.test.mqtt") as scp:
        tg = scp._tg  # noqa:SLF001
        st = S(tg, client_ctx)
        async with AsyncExitStack() as ex:
            st.ex = ex  # pylint: disable=attribute-defined-outside-init
            ex.enter_context(mock.patch("time.time", new=tm))
            ex.enter_context(mock.patch("time.monotonic", new=tm))
            logging._startTime = tm()  # noqa:SLF001

            async def run_broker(cfg):
                async with create_broker(config=cfg) as srv:
                    # NB: some services use "async with await …"
                    scope.register(srv)
                    await scope.no_more_dependents()

            async def with_broker(s, *a, **k):
                await scope.service("moat.mqtt.broker", run_broker, broker_cfg)
                s._scope = scope.get()  # noqa:SLF001
                return await s._scoped_serve(*a, **k)  # noqa:SLF001

            args_def = kw.get("args", attrdict())
            for i in range(n):
                name = "test_" + str(i)
                args = kw.get(name, args_def)
                args = combine_dict(
                    args,
                    args_def,
                    {
                        "cfg": {
                            "conn": {"ssl": client_ctx},
                            "server": {
                                "bind_default": {
                                    "host": "127.0.0.1",
                                    "port": i + PORT + 1,
                                    "ssl": server_ctx,
                                },
                                "backend": "mqtt",
                                "mqtt": {"uri": URI},
                            },
                        },
                    },
                    {"cfg": TESTCFG},
                )
                args_def.pop("init", None)
                s = Server(name, **args)
                ex.enter_context(
                    mock.patch.object(s, "_set_tock", new=partial(mock_set_tock, s, s._set_tock)),  # noqa:SLF001
                )
                ex.enter_context(
                    mock.patch.object(s, "_get_host_port", new=partial(mock_get_host_port, st)),
                )
                st.s.append(s)

            evts = []
            for i in range(n):
                if kw.get(f"run_{i}", run):
                    evt = anyio.Event()
                    await scp.spawn_service(with_broker, st.s[i], ready_evt=evt)
                    evts.append(evt)
                else:
                    setattr(st, f"run_{i}", partial(scp.spawn_service, with_broker, st.s[i]))

            for e in evts:
                await e.wait()
            try:
                done = True
                yield st
            finally:
                logger.info("Runtime: %s", clock.current_time())
                tg.cancel_scope.cancel()
    if not done:
        yield None



# noqa:D100
from __future__ import annotations

import anyio
import logging
from contextlib import asynccontextmanager

from moat.util import NotGiven
from moat.mqtt.client import MQTTClient

from . import Backend

logger = logging.getLogger(__name__)


class MqttMessage:  # noqa:D101
    def __init__(self, topic, payload):
        self.topic = topic
        self.payload = payload


class MqttBackend(Backend):
    "MoaT-KV's MQTT backend."

    client = None

    @asynccontextmanager
    async def connect(self, *a, **kw):  # noqa:D102
        codec = kw.pop("codec", NotGiven)
        C = MQTTClient(self._tg, codec=codec)
        try:
            await C.connect(*a, **kw)
            self.client = C
            yield self
        finally:
            self.client = None
            with anyio.CancelScope(shield=True):
                await self.aclose()
                await C.disconnect()

    @asynccontextmanager
    async def monitor(self, *topic, codec=NotGiven):  # noqa:D102
        topic = "/".join(str(x) for x in topic)
        logger.info("Monitor %s start", topic)
        try:
            async with self.client.subscription(topic, codec=codec) as sub:

                async def sub_get(sub):
                    async for msg in sub:
                        yield MqttMessage(msg.topic.split("/"), msg.data)

                yield sub_get(sub)
        except anyio.get_cancelled_exc_class():
            raise
        except BaseException:
            logger.exception("Monitor %s end", topic)
            raise
        else:
            logger.info("Monitor %s end", topic)

    def send(self, *topic, payload, **kw):  # pylint: disable=invalid-overridden-method
        """
        Send this payload to this topic.
        """
        # client.publish is also async, pass-thru
        return self.client.publish("/".join(str(x) for x in topic), message=payload, **kw)


@asynccontextmanager
async def connect(**kw):  # noqa:D103
    async with anyio.create_task_group() as tg:
        c = MqttBackend(tg)
        async with c.connect(**kw):
            yield c



# command line interface  # noqa:D100
from __future__ import annotations

import sys

import asyncclick as click

from moat.util import P, Path, yprint
from moat.kv.data import add_dates
from moat.kv.errors import ErrorRoot


@click.group()  # pylint: disable=undefined-variable
@click.pass_obj
async def cli(obj):
    """Manage error records in MoaT-KV."""
    obj.err = await ErrorRoot.as_handler(obj.client)


@cli.command()
@click.option("-s", "--subsys", help="Subsystem to access")
@click.argument("path", nargs=1)
@click.pass_obj
async def resolve(obj, path, subsys):
    """
    Mark an error as resolved.
    """
    path = P(path)
    if subsys:
        e = obj.err.get_error_record(subsys, path)
    else:
        e = obj.err.follow(path)
    if e.resolved:
        print("Already resolved.", file=sys.stderr)
        return
    await e.resolve()


@cli.command()
@click.option("-n", "--node", help="add details from this node")
@click.option("-s", "--subsystem", "subsys", help="only show errors from this subsystem")
@click.option("-r", "--resolved", is_flag=True, help="only resolved errors")
@click.option("-v", "--verbose", count=True, help="add per-node details (-vv for traces)")
@click.option("-a", "--all-errors", is_flag=True, help="add details from all nodes")
@click.option("-d", "--as-dict", default=None, help="Dump a list of all open (or resolved) error.")
@click.option("-p", "--path", default=":", help="only show errors below this subpath")
@click.pass_obj
async def dump(obj, as_dict, path, node, all_errors, verbose, resolved, subsys):
    """Dump error entries."""
    path = P(path)
    path_ = obj.cfg["errors"].prefix
    d = 2

    async def one(r):
        nonlocal y
        val = r.value
        if subsys and val.get("subsystem", "") != subsys:
            return
        if path and val.get("path", ())[: len(path)] != path:
            return

        if not all_errors and bool(val.get("resolved", False)) != resolved:
            return

        add_dates(val)
        try:
            rp = val.path
            if as_dict:
                del val.path
            elif not isinstance(rp, Path):
                rp = Path.build(rp)
        except AttributeError:
            rp = Path("incomplete") + r.path
            if not as_dict:
                val.path = rp
        if rp[: len(path)] != path:
            return
        rp = rp[len(path) :]
        if node is None:
            val.syspath = r.path
        else:
            val.syspath = Path(node) + r.path

        rn = {}

        def disp(rr):
            if node and rr.path[-1] != node:
                return
            val = rr.value
            add_dates(val)
            if verbose < 2:
                rr.value.pop("trace", None)
            rn[rr.path[-1]] = rr if obj.meta else rr.value

        if verbose:
            rs = await obj.client._request(  # noqa:SLF001
                action="get_tree",
                min_depth=1,
                max_depth=1,
                path=path_ + r.path[-2:],
                iter=True,
                nchain=3 if obj.meta else 0,
            )
            async for rr in rs:
                disp(rr)

        elif node:
            rs = await obj.client.get(path_ + r.path[-2:] | node)
            if "value" in rs:
                disp(rs)

        if rn:
            val["nodes"] = rn

        if as_dict is not None:
            yy = y
            if subsys is None:
                yy = yy.setdefault(val.get("subsystem", "unknown"), {})
            for p in rp:
                yy = yy.setdefault(p, {})
            yy[as_dict] = r if obj.meta else val
        else:
            yy = r if obj.meta else r.value

            yprint([yy], stream=obj.stdout)
            print("---", file=obj.stdout)

    y = {}
    res = None

    if node is None and len(path) == 2 and isinstance(path[-1], int):  # single error?
        r = await obj.client.get(path_ + path, nchain=3 if obj.meta else 0)
        # Mangle a few variables so that the output is still OK
        r.path = path
        node = None

        async def ait(r):
            yield r

        res = ait(r)
        path = ()

    if res is None:
        res = obj.client.get_tree(path_, min_depth=d, max_depth=d, nchain=3 if obj.meta else 0)
    async for r in res:
        await one(r)

    if as_dict is not None:
        yprint(y, stream=obj.stdout)



# noqa:D104
from __future__ import annotations

import contextlib

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

try:
    from importlib.metadata import version

    _version = version("moat.kv")
    _version_tuple = tuple(int(x) for x in _version.split("."))

except Exception:  # pragma: no cover
    _version = "0.0.1"
    _version_tuple = (0, 0, 1)

with contextlib.suppress(NameError):
    del version



"""
An online-updated config store

"""

from __future__ import annotations

try:
    from contextlib import asynccontextmanager
except ImportError:  # pragma: no cover
    from async_generator import asynccontextmanager

import logging

from .exceptions import ServerClosedError, ServerError
from .obj import ClientEntry, ClientRoot

logger = logging.getLogger(__name__)


class ConfigEntry(ClientEntry):  # noqa:D101
    @classmethod
    def child_type(cls, name):  # pragma: no cover  # noqa:D102
        name  # noqa:B018
        logger.warning("Online config sub-entries are ignored")
        return ClientEntry

    async def set_value(self, value):  # noqa:D102
        await self.root.client.config._update(self._name, value)  # noqa:SLF001


class ConfigRoot(ClientRoot):  # noqa:D101
    CFG = "config"

    @classmethod
    def child_type(cls, name):  # noqa:D102
        name  # noqa:B018
        return ConfigEntry

    @asynccontextmanager
    async def run(self):  # noqa:D102
        try:
            async with super().run() as x:
                yield x
        except ServerClosedError:  # pragma: no cover
            pass
        except ServerError:  # pragma: no cover
            logger.exception("No config data")



# command line interface  # noqa:D100
from __future__ import annotations

import asyncclick as click

from moat.util import NotGiven, Path, split_arg, yprint
from moat.kv.auth import gen_auth, loader


@click.group(short_help="Manage authorization")  # pylint: disable=undefined-variable
@click.option("-m", "--method", default=None, help="Affect/use this auth method")
@click.pass_obj
async def cli(obj, method):
    """Manage authorization. Usage: … auth METHOD command…. Use '.' for 'all methods'."""
    a = await obj.client._request(action="auth_info")  # noqa:SLF001
    a = a.get("value", None)
    if a is not None:
        a = a["current"]
    obj.auth_current = a
    if method is None:
        obj.auth = a or (await one_auth(obj))
    else:
        if method == "-":
            method = None
        obj.auth = method


async def enum_auth(obj):
    """List all configured auth types."""
    if obj.get("auth", None) is not None:
        yield obj.auth
        return
    # TODO create a method for this
    res = await obj.client._request(  # noqa:SLF001
        action="enum_internal",
        path=Path("auth"),
        iter=False,
        with_data=False,
        empty=True,
        nchain=0,
    )
    for r in res.result:
        print(r)
        yield r


async def one_auth(obj):
    """Return the current auth method (from the command line or as used by the server)."""
    if obj.get("auth", None) is not None:
        return obj.auth
    auth = None
    async for a in enum_auth(obj):
        if auth is not None:
            raise click.UsageError("You need to set the auth method")
        auth = a
    if auth is None:
        raise click.UsageError("You need to set the auth method")
    return auth


async def enum_typ(obj, kind="user", ident=None, nchain=0):
    """List all known auth entries of a kind."""
    async for auth in enum_auth(obj):
        if ident is not None:
            res = await obj.client._request(  # noqa:SLF001
                action="auth_list",
                typ=auth,
                kind=kind,
                ident=ident,
                iter=False,
                nchain=nchain,
            )
            yield res
        else:
            async with obj.client._stream(  # noqa:SLF001
                action="auth_list",
                typ=auth,
                kind=kind,
                nchain=nchain,
            ) as res:
                async for r in res:
                    yield r


@cli.command("list")
@click.pass_obj
async def list_(obj):
    """List known auth methods"""
    async for auth in enum_auth(obj):
        print(auth, file=obj.stdout)


@cli.command()
@click.option("-s", "--switch", is_flag=True, help="Switch to a different auth method")
@click.pass_obj
async def init(obj, switch):
    """Setup authorization"""
    if obj.auth_current is not None and not switch:
        raise click.UsageError("Authentication is already set up")

    await obj.client._request(action="set_auth_typ", typ=obj.auth)  # noqa:SLF001
    if obj.debug >= 0:
        if obj.auth:
            print("Authorization switched to", obj.auth, file=obj.stdout)
        else:
            print("Authorization turned off.", file=obj.stdout)


@cli.group()
async def user():
    """Manage users."""
    pass


@user.command("list")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Print complete results. Default: just the names",
)
@click.pass_obj  # pylint: disable=function-redefined
async def list_user(obj, verbose):
    """List all users (raw data)."""
    async for r in enum_typ(obj, nchain=obj.meta):
        if obj.meta or verbose:
            if obj.debug < 2:
                del r["seq"]
                del r["tock"]
            yprint(r, stream=obj.stdout)
        else:
            print(r.ident, file=obj.stdout)


@user.command()
@click.argument("ident", nargs=1)
@click.pass_obj
async def get(obj, ident):
    """Retrieve a user (processed)."""
    lv = loader(await one_auth(obj), "user", make=True, server=False)
    if obj.DEBUG:
        lv._length = 16  # noqa:SLF001

    u = await lv.recv(obj.client, ident, _initial=False)
    yprint(u.export(), stream=obj.stdout)


@user.command()
@click.argument("args", nargs=-1)
@click.pass_obj
async def add(obj, args):
    """Add a user."""
    await _add_mod_user(obj, args, None)


@user.command()
@click.option("-n", "--new", is_flag=True, help="New: ignore previous content")
@click.argument("ident", nargs=1)
@click.argument("type", nargs=1)
@click.argument("key", nargs=1)
@click.argument("args", nargs=-1)
@click.pass_obj
async def param(obj, new, ident, type, key, args):  # noqa:A002
    """Set user parameters for auth, conversion, etc."""
    auth = await one_auth(obj)
    u = loader(auth, "user", make=True, server=False)
    if obj.DEBUG:
        u._length = 16  # noqa:SLF001
    # ou = await u.recv(obj.client, ident, _initial=False)  # unused
    res = await obj.client._request(  # noqa:SLF001
        action="get_internal",
        path=("auth", auth, "user", ident, type),
        iter=False,
        nchain=3,
    )

    kw = res.get("value", NotGiven)
    if new or kw is NotGiven:
        kw = {}
        res.chain = None
    if key == "-":
        if args:
            raise click.UsageError("You can't set params when deleting")
        res = await obj.client._request(  # noqa:SLF001
            action="delete_internal",
            path=("auth", auth, "user", ident, type),
            iter=False,
            chain=res.chain,
        )

    else:
        kw["key"] = key

        for a in args:
            split_arg(a, kw)

        res = await obj.client._request(  # noqa:SLF001
            action="set_internal",
            path=("auth", auth, "user", ident, type),
            iter=False,
            chain=res.chain,
            value=kw,
        )
    if obj.meta:
        # res.ident = ident
        # res.type = type
        yprint(res, stream=obj.stdout)
    else:
        print(ident, type, key, file=obj.stdout)


@user.command()
@click.argument("ident", nargs=1)
@click.argument("args", nargs=-1)
@click.pass_obj
async def mod(obj, ident, args):
    """Change a user."""
    await _add_mod_user(obj, args, ident)


async def _add_mod_user(obj, args, modify):
    auth = await one_auth(obj)
    u = loader(auth, "user", make=True, server=False)
    if obj.DEBUG:
        u._length = 16  # noqa:SLF001
    if modify:
        ou = await u.recv(obj.client, modify, _initial=False)
        kw = ou.export()
    else:
        kw = {}
    for a in args:
        split_arg(a, kw)

    u = u.build(kw, _initial=False)
    if modify is None or u.ident != modify:
        u._chain = None  # new user  # noqa:SLF001
    else:
        u._chain = ou._chain  # noqa:SLF001
    res = await u.send(obj.client)
    if obj.meta:
        res.ident = u.ident
        yprint(res, stream=obj.stdout)
    else:
        print(u.ident, file=obj.stdout)


@user.command(name="auth")
@click.option(
    "-a",
    "--auth",
    type=str,
    default="root",
    help="Auth params. =file or 'type param=value…' Default: root",
)
@click.pass_obj
async def auth_(obj, auth):
    """Test user authorization."""
    user = gen_auth(auth)  # pylint: disable=redefined-outer-name
    await user.auth(obj.client)
    if obj.debug > 0:
        print("OK.", file=obj.stdout)



# command line interface  # noqa:D100
from __future__ import annotations

import asyncclick as click

from moat.kv.server import Server


@click.command(short_help="Run the MoaT-KV server.")  # pylint: disable=undefined-variable
@click.option(
    "-l",
    "--load",
    type=click.Path(readable=True, exists=True, allow_dash=False),
    default=None,
    help="Event log to preload.",
)
@click.option(
    "-s",
    "--save",
    type=click.Path(writable=True, allow_dash=False),
    default=None,
    help="Event log to write to.",
    hidden=True,
)
@click.option(
    "-i",
    "--incremental",
    default=None,
    help="Save incremental changes, not the complete state",
    hidden=True,
)
@click.option(
    "-I",
    "--init",
    default=None,
    help="Initial value to set the root to. Use only when setting up "
    "a cluster for the first time!",
    hidden=True,
)
@click.option(
    "-e",
    "--eval",
    "eval_",
    is_flag=True,
    help="The 'init' value shall be evaluated.",
    hidden=True,
)
@click.option(
    "-a",
    "--auth",
    "--authoritative",
    is_flag=True,
    help="Data in this file is complete: mark anything missing as known even if not.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force 'successful' startup even if data are missing.",
)
@click.argument("name", nargs=1)
@click.argument("nodes", nargs=-1)
@click.pass_obj
async def cli(obj, name, load, save, init, incremental, eval_, auth, force, nodes):
    """
    This command starts a MoaT-KV server.

    One server in your network needs either an initial datum, or a copy of
    a previously-saved MoaT-KV state. Otherwise, no client connections will
    be accepted until synchronization with the other servers in your MoaT-KV
    network is complete.

    This command requires a unique NAME argument. The name identifies this
    server on the network. Never start two servers with the same name!

    You can force the server to fetch its data from a specific node, in
    case some data are corrupted. (This should never be necessary.)
    """

    kw = {}
    if eval_:
        kw["init"] = eval(init)  # pylint: disable=eval-used
    elif init == "-":
        kw["init"] = None
    elif init is not None:
        kw["init"] = init

    from moat.link.announce import as_service  # noqa: PLC0415

    if load and nodes:
        raise click.UsageError("Either read from a file or fetch from a node. Not both.")
    if auth and force:
        raise click.UsageError("Using both '-a' and '-f' is redundant. Choose one.")

    async with as_service(obj) as evt:
        s = Server(name, cfg=obj.cfg["kv"], **kw)
        if load is not None:
            await s.load(path=load, local=True, authoritative=auth)
        if nodes:
            await s.fetch_data(nodes, authoritative=auth)

        await s.serve(log_path=save, log_inc=incremental, force=force, ready_evt=evt)
